import re
import string
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, IntEnum, StrEnum, auto
from logging import Logger
from typing import Any, Final, TypeVar

from .datetime_pomes import TZ_LOCAL
from .env_pomes import APP_PREFIX, env_get_str, env_get_enum
from .str_pomes import (
    str_as_list, str_sanitize, str_find_char, str_find_whitespace
)
# define a TypeVar constrained to IntEnum or StrEnum
IntStrEnum = TypeVar("IntStrEnum", bound=IntEnum | StrEnum)


class MsgLang(StrEnum):
    """
    Possible languages for error reporting.
    """
    EN = auto()
    PT = auto()


VALIDATION_MSG_LANGUAGE: Final[MsgLang] = env_get_enum(key=f"{APP_PREFIX}_VALIDATION_MSG_LANGUAGE",
                                                       enum_class=MsgLang,
                                                       use_names=False,
                                                       def_value=MsgLang.EN)
VALIDATION_MSG_PREFIX: Final[str] = env_get_str(key=f"{APP_PREFIX}_VALIDATION_MSG_PREFIX",
                                                def_value=APP_PREFIX)
CRON_REGEX: Final[re.Pattern] = re.compile(
     r"^"                                                                                   # start of string
     r"((\*|([0-5]?\d)(-([0-5]?\d))?(/\d+)?)(,(\*|([0-5]?\d)(-([0-5]?\d))?(/\d+)?))*)\s+"   # minute
     r"((\*|([01]?\d|2[0-3])(-([01]?\d|2[0-3]))?(/\d+)?)"
     r"(,(\*|([01]?\d|2[0-3])(-([01]?\d|2[0-3]))?(/\d+)?))*)\s+"                            # hour
     r"((\*|([1-9]|[12]\d|3[01])(-([1-9]|[12]\d|3[01]))?(/\d+)?)"
     r"(,(\*|([1-9]|[12]\d|3[01])(-([1-9]|[12]\d|3[01]))?(/\d+)?))*)\s+"                    # day of month
     r"((\*|(0?[1-9]|1[0-2])(-(0?[1-9]|1[0-2]))?(/\d+)?)"
     r"(,(\*|(0?[1-9]|1[0-2])(-(0?[1-9]|1[0-2]))?(/\d+)?))*)\s+"                            # month
     r"((\*|0-7?(/\d+)?)(,(\*|0-7?(/\d+)?))*)"                                              # day of week
     r"$"                                                                                   # end of string
)
CRON_EMAIL: Final[re.Pattern] = re.compile(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$")


def validate_value(attr: str,
                   value: str | float | Decimal,
                   min_value: int = None,
                   max_value: int = None,
                   values: list = None,
                   ignore_case: bool = False,
                   required: bool = False) -> str | None:
    """
    Validate *value* according to value, range, or membership in *values*, as specified.

    :param attr: the name of the attribute
    :param value: the value to be validated
    :param min_value: if *value* is a string, specifies its minimum length; otherwise, specifies its minimum value
    :param max_value: if *value* is a string, specifies its maximum length; otherwise, specifies its maximum value
    :param values: if provided, requires *val* to be contained therein
    :param ignore_case: specifies whether to ignore capitalization when handling string values
    :param required:  requires *value* to be specified
    :return: *None* if *value* passes validation, or the corresponding error message otherwise
    """
    # initialize the return variable
    result: str | None = None

    if value is None or value == "":
        if isinstance(required, bool) and required:
            # 121: Required attribute
            result = validate_format_error(121,
                                           f"@{attr}")
    elif isinstance(values, list):
        val: str | float | Decimal = value
        vals: list = values
        if ignore_case:
            val = value.lower() if isinstance(value, str) else value
            vals = [v.lower() if isinstance(v, str) else v for v in values]
        if val not in vals:
            length: int = len(values)
            if length == 1:
                # 149: Invalid value {}: must be {}
                result = validate_format_error(149,
                                               value,
                                               values[0],
                                               f"@{attr}")
            else:
                # 150: Invalid value {}: must be one of {}
                result = validate_format_error(150,
                                               value,
                                               values[:length],
                                               f"@{attr}")
    elif isinstance(value, str):
        length: int = len(value)
        if min_value is not None and max_value == min_value and length != min_value:
            # 146: Invalid value {}: length must be {}
            result = validate_format_error(156,
                                           value,
                                           min_value,
                                           f"@{attr}")
        elif max_value is not None and max_value < length:
            # 148: Invalid value {}: length longer than {}
            result = validate_format_error(148,
                                           value,
                                           max_value,
                                           f"@{attr}")
        elif min_value is not None and length < min_value:
            # 147: Invalid value {}: length shorter than {}
            result = validate_format_error(147,
                                           value,
                                           min_value,
                                           f"@{attr}")
    elif ((min_value is not None and value < min_value) or
          (max_value is not None and value > max_value)):
        if min_value is not None and max_value is not None:
            # 151: Invalid value {}: must be in the range {}
            result = validate_format_error(151,
                                           value,
                                           [min_value, max_value],
                                           f"@{attr}")
        elif min_value is not None:
            # 144: Invalid value {}: must be greater than {}
            result = validate_format_error(144,
                                           value,
                                           min_value,
                                           f"@{attr}")
        else:
            # 143: Invalid value {}: must be less than {}
            result = validate_format_error(143,
                                           value,
                                           max_value,
                                           f"@{attr}")
    return result


def validate_bool(source: dict[str, Any],
                  attr: str,
                  default: bool = None,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> bool | None:
    """
    Validate the boolean value associated with *attr* in *source*.

    If provided, this value must be on of:
        - a *bool*
        - the integer *1* or *0*
        - the string *1*, *t*, or *true*, case disregarded
        - the string *0*, *f*, or *false*, case disregarded

    :param source: *dict* containing the value to be validated
    :param attr: the name of the attribute whose value is being validated
    :param default: default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: bool | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # retrieve the value
    value = source.get(suffix)

    # validate it
    if value is None or value == "":
        if default is not None:
            value = default
        elif required:
            # 121: Required attribute
            stat = validate_format_error(121,
                                         f"@{attr}")
    elif isinstance(value, str):
        if value.lower() in ["1", "t", "true"]:
            value = True
        elif value.lower() in ["0", "f", "false"]:
            value = False
        else:
            # 152: Invalid value {}: must be type {}
            stat = validate_format_error(152,
                                         value,
                                         "bool",
                                         f"@{attr}")
    # bool is subtype of int
    elif isinstance(value, int) and not isinstance(value, bool):
        if value == 1:
            value = True
        elif value == 0:
            value = False
        else:
            # 152: Invalid value {}: must be type {}
            stat = validate_format_error(152,
                                         value,
                                         "bool",
                                         f"@{attr}")
    elif not isinstance(value, bool):
        # 152: Invalid value {}: must be type {}
        stat = validate_format_error(152,
                                     value,
                                     "bool",
                                     f"@{attr}")
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)
    else:
        result = value

    return result


def validate_int(source: dict[str, Any],
                 attr: str,
                 min_val: int = None,
                 max_val: int = None,
                 values: list[int] = None,
                 default: int = None,
                 required: bool = False,
                 errors: list[str] = None,
                 logger: Logger = None) -> int | None:
    """
    Validate the *int* value associated with *attr* in *source*.

    If provided, this value must be a *int*, or a valid string representation of a *int*.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param values: optional list of allowed values
    :param default: optional default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: int | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # retrieve the value
    value: int = source.get(suffix)

    # validate it ('bool' is subtype of 'int')
    if value is None and isinstance(default, int) and not isinstance(default, bool):
        value = default
    elif isinstance(value, str) and value.isnumeric():
        value = int(value)
    elif value is not None and \
            (isinstance(value, bool) or not isinstance(value, int)):
        # 152: Invalid value {}: must be type {}
        stat = validate_format_error(152,
                                     value,
                                     "int",
                                     f"@{attr}")
    if not stat:
        stat = validate_value(attr=attr,
                              value=value,
                              min_value=min_val,
                              max_value=max_val,
                              values=values,
                              required=required)
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)
    else:
        result = value

    return result


def validate_decimal(source: dict[str, Any],
                     attr: str,
                     min_val: float = None,
                     max_val: float = None,
                     required: bool = False,
                     values: list[float | int] = None,
                     default: float = None,
                     errors: list[str] = None,
                     logger: Logger = None) -> Decimal | None:
    """
    Validate the *float* value associated with *attr* in *source*.

    If provided, this value must be a *float*, or a valid string representation of a *float*.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param values: optional list of allowed values
    :param default: optional default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: Decimal | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # retrieve the value
    value: Decimal = source.get(suffix)

    # validate it
    if value is None and isinstance(default, int | float | Decimal):
        value = Decimal(value=default)
    elif (isinstance(value, float) or
          (isinstance(value, int) and not isinstance(value, bool)) or
          (isinstance(value, str) and value.replace(".", "", 1).isnumeric())):
        value = Decimal(value=value)
    elif isinstance(value, bool) or \
            (value is not None and not isinstance(value, int | float | Decimal)):
        # 152: Invalid value {}: must be type {}
        stat = validate_format_error(152,
                                     value,
                                     "decimal",
                                     f"@{attr}")
    if not stat:
        stat = validate_value(attr=attr,
                              value=value,
                              min_value=min_val,
                              max_value=max_val,
                              values=values,
                              required=required)
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)
    else:
        result = value

    return result


def validate_str(source: dict[str, Any],
                 attr: str,
                 min_length: int = None,
                 max_length: int = None,
                 values: list[str] = None,
                 default: str = None,
                 ignore_case: bool = False,
                 required: bool = False,
                 errors: list[str] = None,
                 logger: Logger = None) -> str | None:
    """
    Validate the *str* value associated with *attr* in *source*.

    If provided, this value must be a *str*.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_length: optional minimum length accepted
    :param max_length:  optional maximum length accepted
    :param values: optional list of allowed values
    :param default: optional default value, overrides *required*
    :param ignore_case: specifies whether to ignore capitalization
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: str | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain the value
    value: str | None = source.get(suffix)

    # validate it
    if value is None and isinstance(default, str):
        value = default
    elif value is not None and not isinstance(value, str):
        # 152: Invalid value {}: must be type {}
        stat = validate_format_error(152,
                                     value,
                                     "str",
                                     f"@{attr}")
    else:
        stat = validate_value(attr=attr,
                              value=value,
                              min_value=min_length,
                              max_value=max_length,
                              values=values,
                              ignore_case=ignore_case,
                              required=required)
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)
    else:
        result = value

    return result


def validate_date(source: dict[str, Any],
                  attr: str,
                  day_first: bool = False,
                  default: date = None,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> date | None:
    """
    Validate the *date* value associated with *attr* in *source*.

    If provided, this value must be a *date* or *datetime*, a valid string representation of
    a *date* or *datetime*, or an integer or float encoding a UNIX-style *timestamp*.
    If the value obtained is a *datetime*, its *date* part is returned.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param day_first: indicates that the day precedes the month in the string representing the date
    :param default: optional default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # import needed module
    from .datetime_pomes import date_parse

    # initialize the return variable
    result: date | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain the value
    value: str = source.get(suffix)

    # validate it
    if value:
        if isinstance(value, datetime):
            result = value.date()
        elif isinstance(value, date):
            result = value
        elif isinstance(value, int | float):
            result = datetime.fromtimestamp(value).date()
        else:
            result = date_parse(dt_str=value,
                                dayfirst=day_first)
        if not result:
            # 141: Invalid value {}
            stat = validate_format_error(141,
                                         value,
                                         f"@{attr}")
        elif result > datetime.now(tz=TZ_LOCAL).date():
            # 153: Invalid value {}: date is later than the current date
            stat = validate_format_error(153,
                                         value,
                                         f"@{attr}")
    elif isinstance(default, date):
        result = default
    elif isinstance(required, bool) and required:
        # 121: Required attribute
        stat = validate_format_error(121,
                                     f"@{attr}")
    if stat:
        result = None
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)

    return result


def validate_datetime(source: dict[str, Any],
                      attr: str,
                      day_first: bool = True,
                      default: datetime = None,
                      required: bool = False,
                      errors: list[str] = None,
                      logger: Logger = None) -> datetime | None:
    """
    Validate the *datetime* value associated with *attr* in *source*.

    If provided, this value must be a *date* or *datetime*, a valid string representation of
    a *date* or *datetime*, or an integer or float encoding a UNIX-style timestamp.
    If the value obtained is a *date*, it is converted to a *datetime* by appending *00:00:00*.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param day_first: indicates that the day precedes the month in the string representing the date
    :param default: optional default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # import needed module
    from .datetime_pomes import datetime_parse

    # initialize the return variable
    result: datetime | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain and validate the value
    value: str = source.get(suffix)
    if value:
        if isinstance(value, datetime):
            result = value
        elif isinstance(value, date):
            result = datetime.combine(result, time())
        elif isinstance(value, int | float):
            result = datetime.fromtimestamp(value)
        else:
            result = datetime_parse(dt_str=value,
                                    dayfirst=day_first)
        if not result:
            # 141: Invalid value {}
            stat = validate_format_error(141,
                                         value,
                                         f"@{attr}")
        elif result > datetime.now(tz=TZ_LOCAL):
            # 153: Invalid value {}: date is later than the current date
            stat = validate_format_error(153,
                                         value,
                                         f"@{attr}")
    elif isinstance(default, datetime):
        result = default
    elif isinstance(required, bool) and required:
        # 121: Required attribute
        stat = validate_format_error(121,
                                     f"@{attr}")
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)

    return result


def validate_enum(source: dict[str, Any],
                  attr: str,
                  enum_class: type[IntEnum | StrEnum],
                  use_names: bool = False,
                  values: list[IntEnum | StrEnum | str | int | Decimal] = None,
                  default: IntEnum | StrEnum | str | int | Decimal = None,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> IntStrEnum:
    """
    Validate the *enum* value associated with *attr* in *source*.

    If provided, this value must be a name or a value corresponding to an instance of a subclass of *enum_class*.
    The only accepted values for *enum_class* are subclasses of  *StrEnum* or *IntEnum*.
    The parameter *use_names* determines whether the names of the elements in the *enum_class* should be used,
    ignoring capitalization. If not specified, the values of the elements are used (the default).
    If *values* is specified, the value obtained is checked for occurrence therein.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param enum_class: the *enum* class to consider (must be a subclass of *IntEnum* or *StrEnum*)
    :param use_names: specifies whether *enum*'s names should be used (defaults to using *enum*'s values)
    :param default: optional default value, overrides *required*
    :param values: optional list of allowed values (defaults to all elements of *enum_class*)
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value as an instance of *enum_class*, or *None* if validation failed
    """
    # initialize the return variable
    result: IntEnum | StrEnum | None = None

    if use_names:
        pos: int = attr.rfind(".") + 1
        suffix: str = attr[pos:]
        value: Any = source.get(suffix)
        if isinstance(value, Enum):
            source = source.copy()
            source[attr] = value.name
        vals: list[str | int | Decimal] = [v.name if isinstance(v, Enum) else v
                                           for v in (values or enum_class._member_names_)]
        name: str = validate_str(source=source,
                                 attr=attr,
                                 values=vals,
                                 default=default.name if isinstance(default, Enum) else default,
                                 ignore_case=True,
                                 required=required,
                                 errors=errors,
                                 logger=logger)
        if name:
            for e in enum_class:
                if e.name.lower() == name.lower():
                    result = e
                    break
    else:
        value: Any = None
        vals: list[str | int | Decimal] = [v.value if isinstance(v, Enum) else v
                                           for v in (values or enum_class)]
        if issubclass(enum_class, StrEnum):
            value: str = validate_str(source=source,
                                      attr=attr,
                                      values=vals,
                                      default=default.value if isinstance(default, Enum) else default,
                                      required=required,
                                      errors=errors,
                                      logger=logger)
        elif issubclass(enum_class, IntEnum):
            value: int = validate_int(source=source,
                                      attr=attr,
                                      values=vals,
                                      default=default.value if isinstance(default, Enum) else default,
                                      required=required,
                                      errors=errors,
                                      logger=logger)
        if value:
            result = enum_class(value)

    return result


def validate_email(source: dict[str, Any],
                   attr: str,
                   default: str = None,
                   required: bool = False,
                   errors: list[str] = None,
                   logger: Logger = None) -> str | None:
    """
    Validate the email value associated with *attr* in *source*.

    If provided, this value must be a syntactically correct email.

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param default: optional default value, overrides *required*
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: str | None = None

    value: str = validate_str(source=source,
                              attr=attr,
                              default=default,
                              required=required,
                              errors=errors,
                              logger=logger)
    if value:
        if CRON_EMAIL.match(value):
            result = value
        elif isinstance(errors, list):
            # 141: Invalid value {}
            errors.append(validate_format_error(141,
                                                value,
                                                f"@{attr}"))
    elif isinstance(required, bool) and required and isinstance(errors, list):
        # 121: Required attribute
        errors.append(validate_format_error(121,
                                            f"@{attr}"))
    return result


def validate_pwd(source: dict[str, Any],
                 attr: str,
                 required: bool = False,
                 errors: list[str] = None,
                 logger: Logger = None) -> str | None:
    r"""
    Validate the password value associated with *attr* in *source*.

    If provided, this value must abide by the following rules:
      - length >= 8
      - at least 1 digit (in range [0-9])
      - at least 1 lower-case letter (in range [a-z])
      - at least 1 uppercase letter (in renge [A-Z])
      - at least 1 special character (#$%&"'()*+,-./:;<=>?@[\]^_`{|}~!)

    :param source: *dict* containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated value, or *None* if validation failed
    """
    # initialize the return variable
    result: str | None = None

    value: str = validate_str(source=source,
                              attr=attr,
                              required=required,
                              errors=errors,
                              logger=logger)
    if value:
        if (len(value) >= 8 and
            str_find_char(source=value,
                          chars=string.digits) >= 0 and
            str_find_char(source=value,
                          chars=string.ascii_uppercase) >= 0 and
            str_find_char(source=value,
                          chars=string.ascii_lowercase) >= 0 and
            str_find_char(source=value,
                          chars=string.punctuation) >= 0):
            result = value
        elif isinstance(errors, list):
            # 237: Value {} does not meet the formation rules
            errors.append(validate_format_error(237,
                                                value,
                                                f"@{attr}"))
    elif isinstance(required, bool) and required and isinstance(errors, list):
        # 121: Required attribute
        errors.append(validate_format_error(121,
                                            f"@{attr}"))
    return result


def validate_cron(source: dict[str, Any],
                  attr: str,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> str | None:
    r"""
    Validate the *CRON* expression associated with *attr* in *source*.

    A valid *CRON* expression has the syntax *<min> <hour> <day> <month> <day-of-week>*, and can include:
      - numbers (e.g. '5')
      - ranges (e.g. '1-5')
      - lists (e.g. '1,2,3')
      - steps (e.g. '*/15')
      - wildcards ('*')

    The following meta-expressions may be substituted for their equivalent CRON expressions:
      - @annually: "0 0 1 1 *"  (on January 1st, at 00h00)
      - @daily:    "1 0 * * *"  (daily, at 00h01)
      - @hourly:   "0 * * * *"  (every hour, at minute 0 - ??h00)
      - @midnight: "0 0 * * *"  (daily, at 00h00)
      - @monthly:  "0 0 1 * *"  (on the first day of the month, at 00h00)
      - @reboot:   "1 0 * * *"  (same as @daily)
      - @weekly:   "0 0 * * 0"  (on Sundays, at 00h00)
      - @yearly:   "0 0 1 1 *"  (same as @anually)

    :param source: *dict* containing the expression to be validated
    :param attr: the attribute associated with the expression to be validated
    :param required: specifies whether a value must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the validated expression, or *None* if validation failed
    """
    # initialize the return variable
    result: str | None = None

    expr: str = validate_str(source=source,
                             attr=attr,
                             required=required,
                             errors=errors,
                             logger=logger)
    if expr:
        match expr:
            case "@annually" | "@yearly":
                result = "0 0 1 1 *"                      # on January 1st, at 00h00
            case "@daily" | "@reboot":
                result = "1 0 * * *"                      # daily, at 00h01
            case "@hourly":
                result = "0 * * * *"                      # every hour, at minute 0 (??h00)
            case "@midnight":
                result = "0 0 * * *"                      # daily, at 00h00
            case "@monthly":
                result = "0 0 1 * *"                      # on the first day of the month, at 00h00
            case "@weekly":
                result = "0 0 * * 0"                      # on Sundays, at 00h00
            case _:
                if CRON_REGEX.match(expr):
                    result = expr
                elif isinstance(errors, list):
                    # 213: Invalid CRON expression {}
                    errors.append(validate_format_error(238,
                                                        expr,
                                                        f"@{attr}"))
    elif isinstance(required, bool) and required and isinstance(errors, list):
        # 121: Required attribute
        errors.append(validate_format_error(121,
                                            f"@{attr}"))
    return result


def validate_ints(source: dict[str, Any],
                  attr: str,
                  sep: str = ",",
                  min_val: int = None,
                  max_val: int = None,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> list[int] | None:
    """
    Validate the list of *int* values associated with *attr* in *source*.

    If provided as a string, the elements therein must be valid string representations of *ints*, separated
    by *sep* (which defaults to a *comma*). Note that an empty string os an empty list are acceptable values,
    yielding an empty list of *ints*.

    :param source: *dict* containing the list of values to be validated
    :param attr: the attribute associated with the list of values to be validated
    :param sep: the separator in the list of values
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param required: whether the list of values must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the list of validated values, or *None* if validation failed or not required and no values found
    """
    # initialize the return variable
    result: list[int] | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain the values
    values: list = source.get(suffix)
    if values:
        if isinstance(values, str):
            values = str_as_list(source=values,
                                 sep=sep)
        if isinstance(values, list):
            ints: list[int] = []
            if len(values) > 0:
                for inx, value in enumerate(values):
                    if (isinstance(value, str) and value.isdigit()) or \
                            (isinstance(value, int) and not isinstance(value, bool)):
                        ints.append(int(value))
                        stat = validate_value(attr=f"@{attr}[{inx+1}]",
                                              value=int(value),
                                              min_value=min_val,
                                              max_value=max_val)
                    else:
                        # 152: Invalid value {}: must be type {}
                        stat = validate_format_error(152,
                                                     value,
                                                     "int",
                                                     f"@{attr}[{inx+1}]")
                    if stat:
                        break
            if not stat:
                result = ints
        else:
            # 152: Invalid value {}: must be type {}
            stat = validate_format_error(152,
                                         values,
                                         "list",
                                         f"@{attr}")

    if isinstance(required, bool) and required and not stat and result is None:
        # 121: Required attribute
        stat = validate_format_error(121,
                                     f"@{attr}")
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)

    return result


def validate_strs(source: dict[str, Any],
                  attr: str,
                  sep: str = ",",
                  min_length: int = None,
                  max_length: int = None,
                  required: bool = False,
                  errors: list[str] = None,
                  logger: Logger = None) -> list[str] | None:
    """
    Validate the list of *str* values associated with *attr* in *source*.

    If provided as a string, the elements therein must be separated by *sep* (which defaults to a *comma*).
    Note that an empty string os an empty list are acceptable values, yielding an empty list of *strs*.

    :param source: *dict* containing the list of values to be validated
    :param attr: the attribute associated with the list of values to be validated
    :param sep: the separator in the list of values
    :param min_length: optional minimum length accepted
    :param max_length:  optional maximum length accepted
    :param required: whether the list of values must be provided
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the list of validated values, or *None* if validation failed or not required and no values found
    """
    # initialize the return variable
    result: list[str] | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain the values
    values: list = source.get(suffix)
    if values:
        if isinstance(values, str):
            values = str_as_list(source=values,
                                 sep=sep)
        if isinstance(values, list):
            strs: list[str] = []
            if len(values) > 0:
                for inx, value in enumerate(values):
                    strs.append(value)
                    if isinstance(value, str):
                        stat = validate_value(attr=f"@{attr}[{inx+1}]",
                                              value=value,
                                              min_value=min_length,
                                              max_value=max_length)
                    else:
                        # 152: Invalid value {}: must be type {}
                        stat = validate_format_error(152,
                                                     value,
                                                     "str",
                                                     f"@{attr}[{inx+1}]")
                    if stat:
                        break
            if not stat:
                result = strs
        else:
            # 152: Invalid value {}: must be type {}
            stat = validate_format_error(152,
                                         values,
                                         "list",
                                         f"@{attr}")

    if required and not stat and result is None:
        # 121: Required attribute
        stat = validate_format_error(121,
                                     f"@{attr}")
    if stat:
        if logger:
            logger.error(msg=stat)
        if isinstance(errors, list):
            errors.append(stat)

    return result


def validate_format_error(error_id: int,
                          *args: Any,
                          **kwargs: dict) -> str:
    """
    Format and return the error message identified by *err_id* in the standard messages list.

    The message is built from the message element in the standard messages list, identified by *error_id*.
    The occurrences of *{}* in the element are sequentially replaced by the given *args*.
    When replacing, an instance of *args* is surrounded by single quotes if it contains no blankspaces.
    The element in *args* prefixed with *@*, if present, is appended to the end of the message.

    Optional custom language and prefix, replacing those defined respectively by the environment variables
    *VALIDATION_MSG_LANGUAGE* and *VALIDATION_MSG_PREFIX*, may be provided in *kwargs*, with the corresponding
    keys *msg_lang*, and *msg_prefix*.

    Suppose this function is invoked with:
      - *error_id*: 147 (defined as: Invalid value {}: length shorter than {})
      - *args*: 'my_value', 10, '@my_attr'
    The formatted error message will be (*<VMP>* is the validation message prefix):
      - <VMP>147: Invalid value 'my_value': length shorter than 10 @my_attr

    :param error_id: the identification of the message element
    :param args: optional non-keyworded arguments to format the error message with
    :param kwargs: optional keyworded arguments to define language and prefix
    :return: the formatted error message
    """
    # obtain definitions for prefix and language
    msg_prefix: str = kwargs.get("msg_prefix") if "msg_prefix" in kwargs else VALIDATION_MSG_PREFIX
    msg_lang: MsgLang = validate_enum(source=kwargs or {},
                                      attr="msg_lang",
                                      enum_class=MsgLang,
                                      default=VALIDATION_MSG_LANGUAGE)
    # retrieve the messages list
    match msg_lang:
        case MsgLang.EN:
            from .validation_msgs import _ERR_MSGS_EN
            err_msgs = _ERR_MSGS_EN
        case MsgLang.PT:
            from .validation_msgs import _ERR_MSGS_PT
            err_msgs = _ERR_MSGS_PT
        case _:
            err_msgs = {}

    # initialize the return variable
    result: str = ""
    if msg_prefix:
        result += msg_prefix + str(error_id) + ": "
    result += err_msgs.get(error_id) or ""

    # apply the provided arguments
    for arg in args:
        if arg is None:
            pos1: int = result.find(": {}")
            pos2: int = result.find(" {}")
            if pos1 < 0 or pos2 < pos1:
                result = result.replace(" {}", "", 1)
            else:
                result = result.replace(": {}", "", 1)
        elif not result or (isinstance(arg, str) and arg.startswith("@")):
            result += " " + arg
        elif isinstance(arg, str) and arg.find(" ") > 0:
            result = result.replace("{}", arg, 1)
        else:
            result = result.replace("{}", f"'{arg}'", 1)

    return result


def validate_format_errors(errors: list[str],
                           **kwargs: dict) -> list[dict[str, str]]:
    """
    Build and return a list of *dicts* from the list of errors in *errors*.

    Each element in *errors* is encoded as a *dict*.
    This list is tipically used in a returning *JSON* string.

    Optional custom language and prefix, replacing those defined respectively by the environment variables
    *VALIDATION_MSG_LANGUAGE* and *VALIDATION_MSG_PREFIX*, may be provided in *kwargs*, with the corresponding
    keys *msg_lang*, and *msg_prefix*.

    Suppose *errors* contains (*<VMP>* is the validation message prefix):
      - <VMP>147: Invalid value 'my_value': length shorter than 10 @my_attr
    The returned list will contain the *dict*:
      - {
      -   "attribute": "my_attr",
      -   "code": <VMP>147,
      -   "description": Invalid value 'my_value': length shorter than 10
      - }

    :param errors: the list of errors to build the list of *dicts* with
    :param kwargs: optional keyworded arguments to define formatting language and prefix
    :return: the built list of *dicts*
    """
    # obtain definitions for prefix and language
    msg_prefix: str = kwargs.get("msg_prefix") if "msg_prefix" in kwargs else VALIDATION_MSG_PREFIX
    msg_lang: MsgLang = validate_enum(source=kwargs or {},
                                      attr="msg_lang",
                                      enum_class=MsgLang,
                                      default=VALIDATION_MSG_LANGUAGE)
    # initialize the return variable
    result: list[dict[str, str]] = []

    # extract error code, description, and attribute from text
    for error in errors:

        # locate the last indicator for the attribute
        pos = error.rfind("@")

        # is there a whitespace in the attribute's name ?
        if pos > 0 and str_find_whitespace(error[pos:]) > 0:
            # yes, disregard the attribute
            pos = -1

        # was the attribute's name found ?
        if pos == -1:
            # no
            out_error: dict[str, str] = {}
            desc: str = error
        else:
            # yes
            term: str = "attribute" if msg_lang == MsgLang.EN else "atributo"
            out_error: dict[str, str] = {term: error[pos + 1:]}
            desc: str = error[:pos - 1]

        # does the text contain an error code ?
        if msg_prefix and desc.startswith(msg_prefix):
            # yes
            term: str = "code" if msg_lang == MsgLang.EN else "codigo"
            pos: int = desc.find(":")
            out_error[term] = desc[0:pos]
            desc = desc[pos+2:]

        term: str = "description" if msg_lang == MsgLang.EN else "descricao"
        out_error[term] = desc
        result.append(out_error)

    return result


def validate_unformat_errors(errors: list[dict[str, str] | str],
                             **kwargs: dict) -> list[str]:
    """
    Extract and return the list of errors used to build the list of dicts *errors*.

    Optional custom language, replacing the one defined by the environment variable
    *VALIDATION_MSG_LANGUAGE*, may be provided in *kwargs*, with the key *msg_lang*.

    :param errors: the list of dicts to extract the errors from
    :param kwargs: optional keyworded arguments to define formatting language
    :return: the built list of errors
    """
    # obtain definitions for language and prefix
    msg_lang: MsgLang = validate_enum(source=kwargs or {},
                                      attr="msg_lang",
                                      enum_class=MsgLang,
                                      default=VALIDATION_MSG_LANGUAGE)
    # initialize the return variable
    result: list[str] = []

    # define the dictionary keys
    name: str = "code" if msg_lang == MsgLang.EN else "codigo"
    desc: str = "description" if msg_lang == MsgLang.EN else "descricao"

    # traverse the list of dicts
    for error in errors:
        if isinstance(error, dict):
            desc: str = str_sanitize(source=error.get(desc) or "''")
            result.append(f"{error.get(name)}: {desc}")
        else:
            result.append(error)

    return result



from datetime import date, datetime, timedelta
from dateutil import parser
from dateutil.parser import ParserError
from enum import StrEnum
from typing import Final
from zoneinfo import ZoneInfo

from .env_pomes import APP_PREFIX, env_get_str

# HAZARD: requires 'tzdata' package installed to work
TZ_LOCAL: Final[ZoneInfo] = ZoneInfo(key=env_get_str(key=f"{APP_PREFIX}_TZ_LOCAL",
                                                     def_value="America/Sao_Paulo"))


class DateFormat(StrEnum):
    """
    Some useful date formats.
    """
    STD = "%m/%d/%Y"
    COMPACT = "%Y%m%d"
    INV = "%Y-%m-%d"


class DatetimeFormat(StrEnum):
    """
    Some useful datetime formats.
    """
    STD = "%m/%d/%Y %H:%M:%S"
    COMPACT = "%Y%m%d%H%M%S"
    INV = "%Y-%m-%d %H:%M:%S"


def date_reformat(dt_str: str,
                  to_format: str,
                  **kwargs: any) -> str | None:
    """
    Convert the date in *dt_str* to the format especified in *to_format*.

    The argument *dt_str* must represent a valid date, with or without time-of-day information.
    The argument *to_format* must be a valid format for *date* ou *datetime*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst=True*
                - to signal that *day* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *yearfirst=True*
                - to signal that *year* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *fmt=<format>*, to force use of a specific format
    Return *None* se *dt_str* does not contain a valid date.

    :param dt_str: the date to convert
    :param to_format: the format for the conversion
    :param kwargs: optional arguments for the parser in python-dateutil
    :return: the converted date, or *None* if the convertion was not possible
    """
    result: str | None = None
    ts: datetime = parser.parse(timestr=dt_str,
                                **kwargs)
    if ts:
        result = ts.strftime(format=to_format)

    return result


def date_weekday(dt_base: str | date | datetime,
                 weekdays: list[int]) -> date | datetime | None:
    """
    Obtain and return the first date with day-of-week specified in *weekdays*, starting at *dt_base*.

    The parameter *weekdays* lists the accepted days-of-week, indicated as:
      - 0: Monday
      - 1: Tuesday
      - 2: Wednesday
      - 3: Thursday
      - 4: Friday
      - 5: Saturday
      - 6: Sunday

    :param dt_base: the base date, either as *date*, *datetime*, or as a valid string representation of a *date*
    :param weekdays: the list of accepted days-of-week, as 0 (Monday) through 6 (Sunday)
    :return: the first satisfying date starting at *dt_base*, or *None*  if it could not be obtained
    """
    result: date | datetime = date_parse(dt_str=dt_base) if isinstance(dt_base, str) else dt_base
    # 'datetime' is subclass of 'date'
    if isinstance(result, date):
        while result.weekday() not in weekdays:
            result += timedelta(days=1)

    return result


def date_parse(dt_str: str,
               **kwargs: any) -> date | None:
    """
    Obtain and return the *date* object corresponding to *dt_str*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst=True*
                - to signal that *day* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *yearfirst=True*
                - to signal that *year* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *fmt=<format>*, to force use of a specific format
    Return *None* se *dt_str* does not contain a valid date.

    :param dt_str: the date, in a supported format
    :param kwargs: optional arguments for the parser in python-dateutil
    :return: the corresponding date object, or *None* if it could not be obtained
    """
    # declare the return variable
    result: date | None

    try:
        result = parser.parse(timestr=dt_str,
                              **kwargs).date()
    except (TypeError, ParserError, OverflowError):
        result = None

    return result


def datetime_parse(dt_str: str,
                   **kwargs: any) -> datetime | None:
    """
    Obtain and return the *datetime* object corresponding to *dt_str*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst=True*
                - to signal that *day* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *yearfirst=True*
                - to signal that *year* comes before *month* in an ambiguous 3-integer date
                  (e.g. '01/05/09') - defaults to *False*
        -   *fmt=<format>*, to force use of a specific format
    Return *None* if *dt_str* does not contain a valid date.

    :param dt_str: the date, in a supported format
    :param kwargs: optional arguments for the parser in python-dateutil
    :return: the corresponding datetime object, or *None*
    """
    # declare the return variable
    result: datetime | None

    try:
        result = parser.parse(timestr=dt_str,
                              **kwargs)
    except (TypeError, ParserError, OverflowError):
        result = None

    return result


def timestamp_interval(start: date | datetime | float,
                       finish: date | datetime | float) -> tuple[int, int, int, int] | None:
    """
    Calculate the number of hours, minutes, seconds, and microseconds between *start* and *finish*.

    It is assumed that *start* and *finish* refer to the same timezone, and that the latter comes after the former.

    :param start: beginning of the interval, given as *date*, *datetime*, or Unix *timestamp*
    :param finish: end of the interval, given as *date*, *datetime*, or Unix *timestamp*
    :return: the number of hours, minutes, seconds, and microseconds in the interval, or *None* if error
    """
    # initialize the return variable
    result: tuple[int, int, int, int] | None = None

    # obtain the corresponding start datetime
    start_dt: datetime
    if isinstance(start, datetime):
        start_dt = start
    elif isinstance(start, date):
        start_dt = datetime.combine(start, datetime.min.time())
    else:
        start_dt = datetime.fromtimestamp(timestamp=start)

    # obtain the corresponding finish datetime
    finish_dt: datetime
    if isinstance(finish, datetime):
        finish_dt = finish
    elif isinstance(finish, date):
        finish_dt = datetime.combine(finish, datetime.min.time())
    else:
        finish_dt = datetime.fromtimestamp(timestamp=finish)

    # compute the duration parts
    if finish_dt >= start_dt:
        td: timedelta = finish_dt - start_dt
        total_secs: float = td.total_seconds()
        hours: int = int(total_secs // 3600)
        mins: int = int((total_secs % 3600) // 60)
        secs: int = int(total_secs % 60)
        result = (hours, mins, secs, td.microseconds)

    return result


def timestamp_duration(start: date | datetime | float,
                       finish: date | datetime | float,
                       show_microsec: bool = False) -> str | None:
    """
    Calculate and return the interval between *start* and *finish*.

    It is assumed that *start* and *finish* refer to the same timezone, and that the latter comes after the former.
    Elapsed hours, minutes, and seconds are returned. Specify *show_microsec* to include microseconds.

    :param start: beginning of the interval, given as *date*, *datetime*, or Unix *timestamp*
    :param finish: end of the interval, given as *date*, *datetime*, or Unix *timestamp*
    :param show_microsec: include elapsed microseconds
    :return: a formatted string with the duration in hours, minutes, seconds, and microseconds, or *None* if error
    """
    # obtain the interval and returned a formatted string
    interval: tuple[int, int, int, int] = timestamp_interval(start=start,
                                                             finish=finish)
    result: str = f"{interval[0]}h{interval[1]:02d}m{interval[2]:02d}"
    if show_microsec:
        result += f".{interval[3]:06d}"
    result += "s"

    return result



import contextlib
from collections import defaultdict
from datetime import date
from enum import Enum
from pathlib import Path
from typing import Any, Literal


def list_compare(list1: list,
                 list2: list) -> bool:
    """
    Compare the contents of the two lists *list1* e *list2*.

    Return *True* if all the elements in *list1* are also in *list2*, and vice versa, with the same cardinality.
    The input list need not be sorted.

    :param list1: the first list
    :param list2: the second list
    :return: True if the two lists contain the same elements, in the same quantity, in any order
    """
    # initialize the return variable
    result: bool = True

    # are the input parameters lists containing the same number of elements ?
    if isinstance(list1, list) and \
       isinstance(list2, list) and \
       len(list1) == len(list2):
        # yes, verify whether all elements in 'list1' are also in 'list2', in the same quantity
        for elem in list1:
            # is 'elem' in both lists, in the same quantity ?
            if list1.count(elem) != list2.count(elem):
                # no, the lists are not equal
                result = False
                break
    else:
        # no, the lists are not equal
        result = False

    return result


def list_correlate(list_first: list,
                   list_second: list,
                   only_in_first: bool = False,
                   only_in_second: bool = False,
                   in_both: bool = False,
                   is_sorted: bool = False) -> tuple:
    """
    Correlate *list_first* and *list_second* by computing their differences.

    A tuple containing up to three lists is returned, with contents based on the parameters:
      - *only_in_first*: items existing in *list_first* but not in *list_second*
      - *only_in_second*: items existing in *list_second* but not in *list_first*
      - *in_both*: items existing in both lists
    If none of these parameters have been specified, no correlation is carried out and an empty tuple is returned.

    The parameter *is_sorted* indicates that both *list_first* and *list_second* are ascendingly or
    descendingly sorted, and binary search is used in the correlation process.

    Note that, depending on the context, a returning list might be the same object as *list_first*
    or *list_second*. The input lists need not be sorted.

    :param list_first: the first list to consider
    :param list_second: the second list to consider
    :param only_in_first: include list of items existing in *list_first* but not in *list_second*
    :param only_in_second: include list of items existing in *list_second* but not in *list_first*
    :param in_both: include list of items existing in both lists
    :param is_sorted: *list_first* and *list_second* are both ascendingly or descendingly sorted
    :return: a tuple containing up to three lists, resulting from correlating the input lists
    """
    # initialize the return variable
    result: tuple = ()

    if only_in_first or only_in_second or in_both:
        result_first: list = []
        result_second: list = []
        result_both: list = []

        if list_first and list_second:
            if is_sorted:
                if in_both or only_in_first:
                    for item in list_first:
                        if list_bin_search(source=list_second,
                                           item=item) < 0:
                            if only_in_first:
                                result_first.append(item)
                        elif in_both:
                            result_both.append(item)
                if only_in_second:
                    result_second = [item for item in list_second if list_bin_search(source=list_first,
                                                                                     item=item) < 0]
            else:
                if in_both or only_in_first:
                    for item in list_first:
                        if item in list_second:
                            if in_both:
                                result_both.append(item)
                        elif only_in_first:
                            result_first.append(item)
                if only_in_second:
                    result_second = [item for item in list_second if item not in list_first]
        elif list_first and only_in_first:
            result_first = list_first
        elif list_second and only_in_second:
            result_second = list_second

        # assemble the return tuple
        correlations: list[list] = []
        if only_in_first:
            correlations.append(result_first)
        if only_in_second:
            correlations.append(result_second)
        if in_both:
            correlations.append(result_both)
        result = tuple(correlations)

    return result


def list_bin_search(source: list,
                    item: Any) -> int:
    """
    Find the index of *item* in the sorted list *source*, using binary search.

    If *source* is not ascendingly or descendinlgy sorted, the return value is not valid.

    :param source: the sorted list to inspect
    :param item: the item to find
    :return: the position of *item* in *source*, or "-1" if not found
    """
    # initialize the return variable
    result: int = -1

    low: int = 0
    high: int = len(source) - 1
    is_ascending: bool = source[low] < source[high]

    while result < 0 and low <= high:
        mid: int = (low + high) // 2
        if source[mid] == item:
            result = mid
        elif is_ascending:
            if item < source[mid]:
                high = mid - 1
            else:
                low = mid + 1
        elif item > source[mid]:
            high = mid - 1
        else:
            low = mid + 1

    return result


def list_flatten(source: list[str]) -> str:
    """
    Buiild and return a *str* by concatenating with "." the elements in *source*.

    Exemples:
        - ['1', '2', '']     -> '1.2.'
        - ['', 'a', 'b']     -> '.a.b'
        - ['x', '', '', 'y'] -> 'x...y'
        - ['z']              -> 'z'

    :param source: the source list
    :return: the concatenated elements of the source list
    """
    result: str = ""
    for item in source:
        result += "." + item
    return result[1:]


def list_unflatten(source: str) -> list[str]:
    """
    Build and return a *list*, by splitting *source* into its components separated by ".".

    This *list* will contain the extracted components. Exemples:
        - '1.2.'  -> ['1', '2', '']
        - '.a.b'  -> ['', 'a', 'b']
        - 'x...y' -> ['x', '', '', 'y']
        - 'z'     -> ['z']

    :param source: string with components concatenated by "."
    :return: the list of strings containing the concatenated components
    """
    from .str_pomes import str_split_on_mark

    return str_split_on_mark(source, ".")


def list_get_coupled(coupled_elements: list[tuple[str, Any]],
                     primary_element: str,
                     couple_to_same: bool = False) -> Any:
    """
    Retrieve from *coupled_elements*, and return, the element coupled to *primary_element*.

    A coupled element is the second element in the tuple whose first element is *primary_element*.

    If *primary_element* contains an index indication (denoted by *[<pos>]*), this indication is removed.
    This function is used in the transformation of *dicts* (*dict_transform*) and *lists* (*list_transform*),
    from sequences of key pairs.

    If *couple_to_same* is *True* and *primary_element* is missing from *coupled_elements*, then
    it is coupled to itself. Note that *primary_element* may be coupled to *None* in *coupled_elements*,
    in which case it is not considered to be missing.

    :param coupled_elements: list of tuples containing the pairs of elements
    :param primary_element: the primary element
    :param couple_to_same: whether to couple *primary_element* to itself if missing in *coupled_elements*
    :return: the coupled element, or *None* if it is not found and *couple_to_same* is *False*
    """
    # initialize the return variable
    result: Any = None

    # remove the list element indication
    pos1: int = primary_element.find("[")
    while pos1 > 0:
        pos2: int = primary_element.find("]", pos1)
        primary_element = primary_element[:pos1] + primary_element[pos2+1:]
        pos1 = primary_element.find("[")

    # traverse the list of coupled elements
    is_coupled: bool = False
    for coupled_element in coupled_elements:
        # has the primary element been found ?
        if coupled_element[0] == primary_element:
            # yes, return the corresponding coupled element
            result = coupled_element[1]
            is_coupled = True
            break
    if couple_to_same and not is_coupled:
        result = primary_element

    return result


def list_transform(source: list,
                   from_to_keys: list[tuple[str, Any]],
                   prefix_from: str = None,
                   prefix_to: str = None,
                   add_missing: bool = False) -> list:
    """
    Construct a new *list*, recursively transforming elements of type *list* and *dict* found in *source*.

    The conversion of *dict* type elements is documented in the *dict_transform* function.

    The prefixes for the source and destination keys, if defined, have different treatments.
    They are added when searching for values in *Source*, and removed when assigning values
    to the return *dict*.
    If *add_missing* is *True*, the entries in *source* whose keys are missing in *from_to_keys*
    are added to the new *list*.

    :param source: the source *dict* of the values
    :param from_to_keys: the list of tuples containing the source and destination key sequences
    :param prefix_from: prefix to be added to the source keys
    :param prefix_to: prefix to be removed from the target keys
    :param add_missing: whether to add entries in *source* missing in *from_to_keys* (defaults to *False*)
    :return: the new list
    """
    from .dict_pomes import dict_transform

    # initialize the return variable
    result: list = []

    # traverse the source list
    for inx, value in enumerate(source):
        from_keys: str | None = None
        if prefix_from:
            from_keys: str = f"{prefix_from}[{inx}]"

        # obtain the target value
        if isinstance(value, dict):
            to_value: dict = dict_transform(source=value,
                                            from_to_keys=from_to_keys,
                                            prefix_from=from_keys,
                                            prefix_to=prefix_to,
                                            add_missing=add_missing)
        elif isinstance(value, list):
            to_value: list = list_transform(source=value,
                                            from_to_keys=from_to_keys,
                                            prefix_from=from_keys,
                                            prefix_to=prefix_to,
                                            add_missing=add_missing)
        else:
            to_value: Any = value

        # added the value transformed to 'result'
        result.append(to_value)

    return result


def list_elem_with_attr(source: list,
                        attr: str,
                        value: Any) -> Any:
    """
    Locate and return the first element in *source* having an attribute named *attr* with value *value*.

    Values obtained by invoking *get* on the element are also considered. *None* is a valid value for *value*.

    :param source: The list to search for the element
    :param attr: the name of the reference attribute
    :param value: the reference value
    :return: The element in *source* having an attribute *attr* with *value*, or *None*
    """
    # initialize the return variable
    result: Any = None

    # traverse the source list
    for element in source:
        if hasattr(element, attr) and getattr(element, attr) == value:
            result = element
            break
        with contextlib.suppress(Exception):
            if element.get(attr) == value:
                result = element
                break

    return result


def list_elem_starting_with(source: list[str | bytes],
                            prefix: str | bytes,
                            keep_prefix: bool = True) -> str | bytes | None:
    """
    Locate and return the first element in *source* prefixed by *prefix*.

    Retorna *None* se esse elemento não for encontrado.

    :param source: the list to be inspected
    :param prefix: the data prefixing the element to be returned
    :param keep_prefix: defines whether the found element should be returned with the prefix
    :return: the prefixed element, with or without the prefix, or *None* if not found
    """
    # initialize the return variable
    result: str | bytes | None = None

    # traverse the source list
    for elem in source:
        if elem.startswith(prefix):
            if keep_prefix:
                result = elem
            else:
                result = elem[len(prefix)+1:]
            break

    return result


def list_prune_duplicates(target: list,
                          is_sorted: bool = False) -> list:
    """
    Remove duplicate elements from *target*.

    The parameter *is_sorted* indicates that *target* is ascendingly or descendingly sorted.
    In both cases, the original order of the elements in *target* is maintained.
    For convenience, the pruned imput list is returned.

    :param target: the target list
    :param is_sorted: *target* is ascendingly or descendingly sorted
    :return: *target* with its duplicate elements removed
    """
    # mark the boundary of the unique segment in the list
    write_index: int = 1

    # traverse the list
    if is_sorted:
        # remove duplicates by comparing each element with the last unique one
        for read_index in range(1, len(target)):
            if target[read_index] != target[write_index - 1]:
                # add this element to the unique segment in the list
                if write_index != read_index:
                    target[write_index] = target[read_index]
                write_index += 1
    else:
        # remove duplicates by verifying if each element is in the unique segment in the list
        for read_index in range(1, len(target)):
            if target[read_index] not in target[:write_index]:
                # add this element to the unique segment in the list
                if write_index != read_index:
                    target[write_index] = target[read_index]
                write_index += 1

    # delete the remaining tail of the list
    del target[write_index:]

    # return the input list for convenience
    return target


def list_prune_in(target: list,
                  ref: list) -> list:
    """
    Remove from *target* all its elements that are also in *ref*.

    The pruned input list is returned, for convenience.

    :param target: the target list
    :param ref: the reference list
    :return: the target list without the elements also in the reference list
    """
    # initialize the return variable
    result: list = target

    removals: list = [item for item in result if item in ref]
    for item in removals:
        result.remove(item)

    return result


def list_prune_not_in(target: list,
                      ref: list) -> list:
    """
    Remove from *target* all of its elements that are not also in *ref*.

    The pruned input list is returned, for convenience.

    :param target: the target list
    :param ref: the reference list
    :return: the target list without the elements not in the reference list
    """
    # initialize the return variable
    result: list = target

    removals: list = [item for item in result if item not in ref]
    for item in removals:
        result.remove(item)

    return result


def list_jsonify(source: list,
                 jsonify_enums: Literal["names", "values"] | None = "values") -> list:
    """
    Return a new *list* containing the values in *source*, made serializable if necessary.

    Possible transformations:
      - *Enum* is changed to its value or name, as per *jsonify_enums* (defaults to value)
      - *bytes* and *bytearray* are changed with *str()*
      - *date* and *datetime* are changed to their *ISO* representations
      - *Path* is changed to its *POSIX* representation
      - *dict* is recursively *jsonified* with *dict_jsonify()*, unless *jsonify_enums* is *None*
      - *list* is recursively *jsonified* with *list_jsonify()*
      - all other types are left unchanged

    Note that retrieving the original values through a reversal of this process is not deterministic.
    The transformation is recursively carried out, that is, any *dict* or *list* set as a list item
    will be *jsonified* accordingly.

    *HAZARD*: depending on the type of object contained in *source*, the final result may still
    not be fully serializable.

    :param source: the *dict* to be *jasonified*
    :param jsonify_enums: whether to *jsonify* enums, using their values (the default) or names
    :return: a new *jsonified* list
    """
    # initialize the return variable
    result: list = []

    # traverse the input list
    for value in source:
        if isinstance(value, Enum):
            result.append(value.value if jsonify_enums == "values" else value.name)
        elif isinstance(value, bytes | bytearray):
            result.append(str(value))
        elif isinstance(value, date):
            result.append(value.isoformat())
        elif isinstance(value, Path):
            result.append(value.as_posix())
        elif isinstance(value, dict):
            if jsonify_enums:
                from .dict_pomes import dict_jsonify
                result.append(dict_jsonify(source=value,
                                           jsonify_keys=jsonify_enums == "names",
                                           jsonify_values=jsonify_enums == "values"))
        elif isinstance(value, list):
            result.append(list_jsonify(source=value,
                                       jsonify_enums=jsonify_enums))
        else:
            result.append(value)

    return result


def list_hexify(source: list) -> list:
    """
    Return a new *list* containing the values in *source* changed to their hexadecimal representations.

    Possible transformations:
      - *str* is changed with *<value>.encode().hex()*
      - *int* is changed with *float(<value>).hex()*
      - *float*, *bytes*, and *bytearray* are changed using their built-in *hex()* method
      - *date* and *datetime* are changed using their ISO representations
      - *Path* is changed using its POSIX representation
      - *dict* is recursively *hexified* with *dict_hexify()*
      - *list* is recursively *hexified* with *list_hexify()*
      - all other types are left unchanged

    Note that retrieving the original values through a reversal of this process is not deterministic.
    The transformation is recursively carried out, that is, any *dict* or *list* set as a list item
    will be *hexified* accordingly.

    :param source: the list to be *hexified*
    :return: a list with *hexified* values
    """
    from .dict_pomes import dict_hexify

    # initialize the return variable
    result: list = []

    # traverse the input list
    for value in source:
        if isinstance(value, dict):
            dict_hexify(source=value,
                        hexify_keys=False,
                        hexify_values=True)
            result.append(value)
        elif isinstance(value, list):
            result.append(list_hexify(source=value))
        elif isinstance(value, str):
            result.append(value.encode().hex())
        elif isinstance(value, int):
            result.append(float(value).hex())
        elif isinstance(value, float | bytes | bytearray):
            result.append(value.hex())
        elif isinstance(value, Path):
            result.append(value.as_posix().encode().hex())
        elif isinstance(value, date):
            result.append(value.isoformat().encode().hex())
        else:
            result.append(value)

    return result


def list_hierarchize(source: list[list | tuple]) -> list:
    """
    Hierarquize a fully sorted list of tuples or list of lists by aggregating common values at all levels.

    To ilustrate, let us assume *source* is the input list:
    ::
      [
        ('John', 'parent Fred', 'old age', 'indifferent'),
        ('John', 'parent Fred', 'old age', 'unaffected'),
        ('John', 'parent Fred', 'poor health', 'broken'),
        ('John', 'parent Fred', 'poor health', 'constrained'),
        ('John', 'parent Fred', 'poor health', 'dependent'),
        ('John', 'parent Kate', 'happy soul'),
        ('John', 'parent Kate', 'very intelligent'),
        ('Mary', 'child John', 'brown eyes'),
        ('Mary', 'child John', 'red hair'),
        ('Mary', 'child Susan', 'blue eyes'),
        ('Mary', 'child Susan', 'median height'),
        ('Mary', 'sibling Joan', 'charming dude'),
        ('Mary', 'sibling Joan', 'smart girl')
     ]

    The resulting hierarquization would yield the list:
    ::
      [
        ['John',
          ['parent Fred',
            ['old age', ['indifferent', 'unaffected']],
            ['poor health', ['broken', 'constrained', 'dependent']]],
          ['parent Kate', ['happy soul', 'very intelligent']]],
        ['Mary',
          ['child John', ['brown eyes', 'red hair']],
          ['child Susan', ['blue eyes', 'median height']],
          ['sibling Joan', ['charming dude', 'smart girl']]]
      ]

    Notes:
      - the elements in *source* must not contain embedded lists or tuples
      - once an aggregation has been given a value, another aggregation cannot be added to it, such as:

        ('John', 'parent Fred', 'poor health', 'dependent'),

        ('John', 'parent Fred', 'poor health', 'constrained', 'simple'),

        ('John', 'parent Fred', 'poor health', 'constrained', 'complex'),

    :param source: the fully sorted list of tuples or list of lists to be hierarchized
    :return: the hierarchized list

    """
    def add_to_hierarchy(hierarchy: dict,
                         keys: list,
                         value: list | tuple) -> None:
        for key in keys[:-1]:
            hierarchy.setdefault(key, {})
        # if isinstance(hierarchy.get(keys[-1]), dict):
        hierarchy.setdefault(keys[-1], []).append(value)

    def convert_to_list(item: Any) -> list:
        result: list
        if isinstance(item, dict):
            result = []
            for k, v in item.items():
                if isinstance(v, dict):
                    result.append([k, *convert_to_list(item=v)])
                else:
                    result.append([k, v] if len(v) > 1 else [k, *v])
        elif isinstance(item, list):
            result = item
        else:
            result = [item]
        return result

    hierarchy: dict = defaultdict(dict)
    for item in source:
        add_to_hierarchy(hierarchy=hierarchy,
                         keys=item[:-1],
                         value=item[-1])

    return convert_to_list(item=hierarchy)


def list_stringify(source: list) -> str:
    """
    Return a string with the items from *source* listed as *[<i1>, ..., <in>]*.

    The *stringification* is done recursively, with *dict* and *list* as values handled accordingly.

    :param source: the source *list*
    :return: the string listing the items in *source*
    """
    from .dict_pomes import dict_stringify

    # initialize the return variable
    result: str = "["

    if source:
        # traverse the source 'list'
        for item in source:
            if isinstance(item, dict):
                result += f"{dict_stringify(source=item)}"
            elif isinstance(item, list):
                result += f"{list_stringify(source=item)}"
            elif isinstance(item, str):
                result += f"'{item}'"
            else:
                result += f"{item}"
            result += ", "
        result = result[:-2]

    return result + "]"



import inspect
import json
import os
import traceback
from contextlib import suppress
from types import FrameType, TracebackType
from typing import Any


def obj_is_serializable(obj: Any) -> bool:
    """
    Determine if *obj* is serializable.

    :param obj: the reference object
    :return: *True* if serializable, *False* otherwise
    """
    # initialize the return variable
    result: bool = True

    # verify the object
    try:
        json.dumps(obj)
    except (TypeError, OverflowError):
        result = False

    return result


def obj_to_dict(obj: Any,
                omit_private: bool = True) -> dict[str, Any] | list[Any] | Any:
    """
    Convert the generic object *obj* to a *dict*.

    The conversion is done recursively. Attributes for which exceptions are raised on attempt
    to access them are silently omited.

    :param obj: the object to be converted
    :param omit_private: whether to omit private attributes (defaults to *True*)
    :return: the dict obtained from *obj*
    """
    # declare the return variable
    result: dict[str, Any] | list[Any] | Any

    if isinstance(obj, dict):
        result = {str(k): obj_to_dict(obj=v,
                                      omit_private=omit_private) for k, v in obj.items()}
    elif isinstance(obj, list | tuple | set):
        result = [obj_to_dict(obj=item,
                              omit_private=omit_private) for item in obj]
    elif hasattr(obj, "__dict__") or not isinstance(obj, str | int | float | bool | type(None)):
        result = {}
        for attr in dir(obj):
            if not (omit_private and attr.startswith("_")):
                with suppress(Exception):
                    value: Any = getattr(obj, attr)
                    if not callable(value):
                        result[attr] = obj_to_dict(obj=value,
                                                   omit_private=omit_private)
    else:
        result = obj

    return result


def exc_format(exc: Exception,
               exc_info: tuple[type[BaseException], BaseException, TracebackType]) -> str:
    """
    Format the error message resulting from the exception raised in execution time.

    The format to use: <python_module>, <line_number>: <exc_class> - <exc_text>

    :param exc: the exception raised
    :param exc_info: information associated with the exception
    :return: the formatted message
    """
    tback: TracebackType = exc_info[2]
    cls: str = str(exc.__class__)

    # retrieve the execution point where the exception was raised (bottom of the stack)
    tlast: traceback = tback
    while tlast.tb_next:
        tlast = tlast.tb_next

    # retrieve the module name and the line number within the module
    try:
        fname: str = os.path.split(p=tlast.tb_frame.f_code.co_filename)[1]
    except Exception:
        fname: str = "<unknow module>"
    fline: int = tlast.tb_lineno

    return f"{fname}, {fline}, {cls[8:-2]} - {exc}"


def func_get_passed_args(include_defaults: bool = False) -> dict[str, Any]:
    """
    Retrieve the arguments passed to the invoking function, as *key-value* pairs.

    *include_defaults* determines whether to include parameters that were not actually passed,
    but has default values defined.

    :param include_defaults: whether to include parameters with default values that were not actually passed
    :return: a *dict* with the passed parameters as *key-value* pairs
    """
    # get the caller's frame
    frame: FrameType = inspect.currentframe().f_back

    # Get the function object
    func: Any = frame.f_globals[frame.f_code.co_name]

    args_info: inspect.ArgInfo = inspect.getargvalues(frame=frame)
    sig: inspect.Signature = inspect.signature(func)

    bound_args: inspect.BoundArguments
    if include_defaults:
        # bind all arguments, filling in defaults
        bound_args = sig.bind(**args_info.locals)
        bound_args.apply_defaults()
    else:
        # bind only explicitly passed arguments
        bound_args = sig.bind_partial(**args_info.locals)

    return dict(bound_args.arguments)



from typing import Final, Literal

__ERR_MSGS: Final[dict[int, dict[str, str]]] = {
    101: {
        "en": "{}",
        "pt": "{}",
    },
    102: {
        "en": "Unexpected error: {}",
        "pt": "Erro inesperado: {}",
    },
    103: {
        "en": "Invalid operation {}",
        "pt": "Operação {} inválida",
    },
    104: {
        "en": "The operation {} returned the error {}",
        "pt": "A operação {} retornou o erro {}",
    },
    105: {
        "en": "Error invoking service {}: {}",
        "pt": "Erro na invocação do serviço {}: {}",
    },
    106: {
        "en": "No records matching the provided criteria found",
        "pt": "Não foram encontrados registros para os critérios fornecidos",
    },
    107: {
        "en": "No files matching the provided criteria found",
        "pt": "Não foram encontrados arquivos para os critérios fornecidos",
    },
    111: {
        "en": "File {} not found",
        "pt": "Arquivo {} não encontrado",
    },
    112: {
        "en": "Environment variable {} not defined",
        "pt": "Variável de ambiente {} não definida",
    },
    121: {
        "en": "Required attribute",
        "pt": "Atributo obrigatório",
    },
    122: {
        "en": "Attribute is unknown or invalid in this context",
        "pt": "Atributo desconhecido ou inválido para o contexto",
    },
    123: {
        "en": "Attribute not applicable for {}",
        "pt": "Atributo não se aplica a {}",
    },
    124: {
        "en": "Attribute applicable only for {}",
        "pt": "Atributo se aplica apenas a {}",
    },
    125: {
        "en": "A value has not been assigned",
        "pt": "Valor não foi atribuído",
    },
    126: {
        "en": "Value {} cannot be assigned for attributes {} at the same time",
        "pt": "Valor {} não pode ser especificado aos atributos {} ao mesmo tempo",
    },
    127: {
        "en": "Attributes {} cannot be assigned values at the same time",
        "pt": "Atributos {} não podem ter valores especificados ao mesmo tempo",
    },
    141: {
        "en": "Invalid value {}",
        "pt": "Valor {} inválido",
    },
    142: {
        "en": "Invalid value {}: {}",
        "pt": "Valor {} inválido: {}",
    },
    143: {
        "en": "Invalid value {}: must be less than {}",
        "pt": "Valor {} inválido: deve ser menor que {}",
    },
    144: {
        "en": "Invalid value {}: must be greater than {}",
        "pt": "Valor {} inválido: deve ser maior que {}",
    },
    145: {
        "en": "Invalid, inconsistent, or missing arguments",
        "pt": "Argumento(s) inválido(s), inconsistente(s) ou não fornecido(s)",
    },
    146: {
        "en": "Invalid value {}: length must be {}",
        "pt": "Valor {} inválido: comprimento deve ser {}",
    },
    147: {
        "en": "Invalid value {}: length shorter than {}",
        "pt": "Valor {} inválido: comprimento menor que {}",
    },
    148: {
        "en": "Invalid value {}: length longer than {}",
        "pt": "Valor {} inválido: comprimento maior que {}",
    },
    149: {
        "en": "Invalid value {}: must be {}",
        "pt": "Valor {} inválido: deve ser {}",
    },
    150: {
        "en": "Invalid value {}: must be one of {}",
        "pt": "Valor {} inválido: deve ser um de {}",
    },
    151: {
        "en": "Invalid value {}: must be in the range {}",
        "pt": "Valor {} inválido: deve estar no intervalo {}",
    },
    152: {
        "en": "Invalid value {}: must be type {}",
        "pt": "Valor {} inválido: deve ser do tipo {}",
    },
    153: {
        "en": "Invalid value {}: date is later than the current date",
        "pt": "Valor {} inválido: data posterior à data atual",
    },
    154: {
        "en": "Value {} not a valid date",
        "pt": "Valor {} não é uma data válida"
    },
    171: {
        "en": "Error receiving attachment: {}",
        "pt": "Erro no recebimento de documento anexado: {}",
    },
    172: {
        "en": "Invalid attachment type {}: {}",
        "pt": "Tipo de documento anexado {} inválido: {}",
    },
    173: {
        "en": "Unable to receive attachment {}",
        "pt": "Não foi possível receber o documento anexado {}",
    },
    201: {
        "en": "Error accessing the DB in {}: {}",
        "pt": "Erro na interação com o BD em {}: {}",
    },
    202: {
        "en": "No record found on DB in {}, for {}",
        "pt": "Nenhum registro encontrado no BD, em {} para {}",
    },
    203: {
        "en": "Error accessing the object store: {}",
        "pt": "Erro na interação com o armazenador de objetos: {}",
    },
    204: {
        "en": "Unable to retrieve document {} from the object store",
        "pt": "Não foi possível recuperar o documento {} no armazenador de objetos",
    },
    205: {
        "en": "Error accessing the message queue manager: {}",
        "pt": "Erro na interação com o gerenciador de mensagens: {}",
    },
    206: {
        "en": "Record already exists on DB in {}, for {}",
        "pt": "Registro já existe no BD, em {} para {}",
    },
    207: {
        "en": "More than one record exists on DB in {}, for {}",
        "pt": "Mais de um registro existente no BD, em {} para {}",
    },
    211: {
        "en": "Error accessing the job scheduler: {}",
        "pt": "Erro na interação com o gerenciador de tarefas: {}",
    },
    212: {
        "en": "Unable to access the entry point {} for the job {}: {}",
        "pt": "Não foi possível acesso ao ponto de entrada {} para a tarefa {}: {}",
    },
    213: {
        "en": "Invalid CRON expression {}",
        "pt": "Expressão CRON {} inválida",
    },
    221: {
        "en": "Error accessing the SOAP service: {}",
        "pt": "Erro na interação com o serviço SOAP: {}",
    },
    222: {
        "en": "SOAP envelope {} has invalid content: {}",
        "pt": "Envelope SOAP {} com conteúdo inválido: {}",
    },
    231: {
        "en": "Token not provided",
        "pt": "Token não fornecido",
    },
    232: {
        "en": "Invalid token",
        "pt": "Token inválido",
    },
    233: {
        "en": "User not provided",
        "pt": "Usuário não fornecido",
    },
    234: {
        "en": "Invalid user {}",
        "pt": "Usuário {} inválido",
    },
    235: {
        "en": "User/password not provided",
        "pt": "Usuário/senha não fornecido",
    },
    236: {
        "en": "Invalid user/password",
        "pt": "Usuário/senha inválido",
    },
    237: {
        "en": "Value {} does not meet the formation rules",
        "pt": "Valor {} não atende as regras de formação"
    },
    241: {
        "en": "Error accessing the digital signing service at {}: {}",
        "pt": "Erro no acesso ao serviço de assinaturas em {}: {}",
    },
    242: {
        "en": "Failed to verify the digital signature of document {}: {}",
        "pt": "Erro na verificação da assinatura digital do documento {}:{}",
    },
    243: {
        "en": "Failed to verify the digest for document {}: {}",
        "pt": "Erro na verificação do digesto do documento {}: {}",
    },
    244: {
        "en": "Document {} is not digitally signed",
        "pt": "Documento {} não contem assinatura digital",
    },
    245: {
        "en": "Invalid digital signature in document {}: {}",
        "pt": "A assinatura digital no documento {} é inválida: {}",
    },
    246: {
        "en": "Document {} was modified after being digitally signed",
        "pt": "Documento {} modificado após assinatura digital",
    },
    247: {
        "en": "Document {} has invalid digest",
        "pt": "Digesto do documento {} inválido",
    },
    248: {
        "en": "Document {} has already been filed",
        "pt": "Documento {} já cadastrado",
    }
}

_ERR_MSGS_EN: dict[int, str] = {}
for key, value in __ERR_MSGS.items():
    _ERR_MSGS_EN[key] = value["en"]

_ERR_MSGS_PT: dict[int, str] = {}
for key, value in __ERR_MSGS.items():
    _ERR_MSGS_PT[key] = value["pt"]


def validate_set_msgs(msgs: dict[int, str],
                      lang: Literal["en", "pt"] = "en") -> None:
    """
    Set  the standard validation messages list for language *lang* to the coded messages in *msgs*.

    If applicable, this operation should be performed at the start of the application importing this module,
    before any attempt to read from *_ERR_MSGS_EN* or *_ERR_MSGS_PT*.

    :param msgs: list of coded messages to set the standard validation messages to
    :param lang: the reference language
    """
    global _ERR_MSGS_EN, _ERR_MSGS_PT

    match lang:
        case "en":
            _ERR_MSGS_EN = msgs
        case "pt":
            _ERR_MSGS_PT = msgs


def validate_update_msgs(msgs: dict[int, str],
                         lang: Literal["en", "pt"] = "en") -> None:
    """
    Update the messages in the standard validation messages list with *msgs*, for language *lang*.

    If applicable, this operation should be performed at the start of the application importing this module,
    before any attempt to read from *_ERR_MSGS_EN* or *_ERR_MSGS_PT*.

    :param msgs: list of coded messages to update the standard validation messages with
    :param lang: the reference language
    """
    match lang:
        case "en":
            _ERR_MSGS_EN.update(msgs)
        case "pt":
            _ERR_MSGS_PT.update(msgs)



import random
import string
from contextlib import suppress
from datetime import date
from pathlib import Path
from typing import Any


def str_to_hex(source: str) -> str:
    """
    Obtain and return the hex representation of *source*.

    This is an encapsulation of the built-in methods *<str>.encode()* and *<bytes>.hex()*.

    :param source: the input string
    :return: the hex representation of the input string
    :raises AttributeError: *source* is not a string
    :raises UnicodeEncodeError:  a UTF-8 encoding error occurred
    """
    return source.encode().hex()


def str_from_hex(source: str) -> str:
    """
    Obtain and return the original string from its hex representation in *source*.

    This is an encapsulation of the built-in methods *bytes.fromhex()* and *<bytes>.decode()*

    :param source: the hex representation of a string
    :return: the original string
    :raises ValueError: *source* is not a valid hexadecimal string
    :raises TypeError: *source* is not a string
    :raises UnicodeDecodeError: a UTF-8 decoding error occurred
    """
    return bytes.fromhex(source).decode()


def str_as_list(source: str,
                sep: str = ",") -> list[str]:
    """
    Return *source* as a *list*, by splitting its contents separated by *sep*.

    The returned substrings are fully whitespace-trimmed. If *source* is *None*, an empty list is returned.
    If it is an empty *str*, a list containg an empty *str* is returned. If it is not a *str*,
    a list containing itself is returned.

    :param source: the string value to be worked on
    :param sep: the separator (defaults to ",")
    :return: a list built from the contents of *source*, or containing *source* itself, if it is not a string
    """
    # declare the return variable
    result: list[str]

    if isinstance(source, str):
        result = [s.strip() for s in source.split(sep=sep)]
    elif source is None:
        result = []
    else:
        result = [source]

    return result


def str_sanitize(source: str) -> str:
    """
    Clean the given *source* string.

    The sanitization is carried out by:
        - removing backslashes
        - replacing double quotes with single quotes
        - replacing newlines and tabs with whitespace
        - replacing multiple consecutive spaces with a single space

    :param source: the string to be cleaned
    :return: the cleaned string
    """
    cleaned: str = source.replace("\\", "") \
                         .replace('"', "'") \
                         .replace("\n", " ") \
                         .replace("\t", " ")
    return " ".join(cleaned.split())


def str_split_on_mark(source: str,
                      mark: str) -> list[str]:
    """
    Extract from *source* the text segments separated by *mark*, and return them in a *list*.

    The separator itself will not be in the returned list.

    :param source: the string to be inspected
    :param mark: the separator
    :return: the list of text segments extracted
    """
    # inicialize the return variable
    result: list[str] = []

    pos: int = 0
    skip: int = len(mark)
    after: int = source.find(mark)
    while after >= 0:
        result.append(source[pos:after])
        pos = after + skip
        after = source.find(mark, pos)
    if pos < len(source):
        result.append(source[pos:])
    else:
        result.append("")

    return result


def str_find_char(source: str,
                  chars: str) -> int:
    """
    Locate and return the position of the first occurence, in *source*, of a character in *chars*.

    :param source: the string to be inspected
    :param chars: the reference characters
    :return: the position of the first character in *chars*, or *-1* if none was found
    """
    # initialize the return variable
    result: int = -1

    # search for whitespace
    for inx, char in enumerate(source):
        if char in chars:
            result = inx
            break

    return result


def str_find_whitespace(source: str) -> int:
    """
    Locate and return the position of the first occurence of a *whitespace* character in *source*.

    :param source: the string to be inspected
    :return: the position of the first whitespace character, or *-1* if none was found
    """
    # initialize the return variable
    result: int = -1

    # search for whitespace
    for inx, char in enumerate(source):
        if char.isspace():
            result = inx
            break

    return result


def str_between(source: str,
                from_str: str,
                to_str: str) -> str:
    """
    Extract and return the *substring* in *source* located between the delimiters *from_str* and *to_str*.

    :param source: the string to be inspected
    :param from_str: the initial delimiter
    :param to_str: the final delimiter
    :return: the extracted substring, or *None* if no substring was obtained
    """
    # initialize the return variable
    result: str | None = None

    pos1: int = source.find(from_str)
    if pos1 >= 0:
        pos1 += len(from_str)
        pos2: int = source.find(to_str, pos1)
        if pos2 >= pos1:
            result = source[pos1:pos2]

    return result


def str_positional(source: str,
                   keys: tuple[str, ...],
                   values: tuple[str, ...]) -> Any:
    """
    Locate the position of *source* within *keys*, and return the element in the same position in *values*.

    :param source: the source string
    :param keys: the tuple holding the keys to be inspected
    :param values: the tuple holding the positionally corresponding values
    :return: the value positionally corresponding to the source string, or *None* if not found
    """
    # noinspection PyUnusedLocal
    result: Any = None
    with suppress(Exception):
        pos: int = keys.index(source)
        result = values[pos]

    return result


def str_random(size: int,
               chars: str | list[str] = None) -> str:
    """
    Generate and return a random string containing *len* characters.

    If *chars* is  provided, either as a string or as a list of characters, the characters
    therein will be used in the construction of the random string. Otherwise, a concatenation of
    *string.ascii_letters*, *string.digits*, and *string.puctuation* will provide the base characters.

    :param size: the length of the target random string
    :param chars: optional characters to build the random string from (a string or a list of characteres)
    :return: the random string
    """
    # establish the base characters
    if not chars:
        chars: str = string.ascii_letters + string.digits + string.punctuation
    elif isinstance(chars, list):
        chars: str = "".join(chars)

    # generate and return the random string
    # ruff: noqa: S311 - Standard pseudo-random generators are not suitable for cryptographic purposes
    return "".join(random.choice(seq=chars) for _ in range(size))


def str_rreplace(source: str,
                 old: str,
                 new: str,
                 count: int = 1) -> str:
    """
    Replace at most *count* occurrences of substring *old* with string *new* in *source*, in reverse order.

    :param source: the string to have a substring replaced
    :param old: the substring to replace
    :param new: the string replacement
    :param count: the maximum number of replacements (defaults to 1)
    :return: the modified string
    """
    return source[::-1].replace(old[::-1], new[::-1], count)[::-1]


def str_splice(source: str,
               seps: tuple[str, ...]) -> tuple[str, ...]:
    """
    Splice *source* into segments delimited by the ordered list of separators *seps*.

    The number of segments returned is always the number of separators in *seps*, plus 1.
    An individual segment returned can be null or an empty string. If no separators are found,
    the returned tuple will contain *source* as its last element, and *None* as the remaining elements.
    Separators will not be part of their respective segments.

    Separators in *seps* can not be *None* or empty strings. If no separators are provided
    (*seps* itself is an empty list), then the returned tuple will contain *source* as its only element.
    If *source* starts with the separator, then the return tuple's first element will be an empty string.
    If *source* ends with the separator, then the return tuple's last element will be an empty string.

    These examples illustrate the various possibilities (*source* = 'My string to be spliced'):
      - () ===> ('My string to be spliced',)
      - ('My') ===> ('',  'string to be spliced')
      - ('tri') ===> ('My s', 'ng to be spliced')
      - ('iced') ===> ('My string to be spl', '')
      - ('X', 'B') ===> (None, None, 'My string to be spliced')
      - ('M', 'd') ===> ('', 'y string to be splice', '')
      - ('s', 's', 'd') ===> ('My ', 'tring to be ', 'plice', '')
      - ('X', 'ri', 'be') ===> (None, 'My st', 'ng to ', ' spliced')

    :param source: the source string
    :param seps: the ordered list of separators
    :return: tuple with the segments obtained, or *None* if *source* is not a string

    """
    # initialize the return variable
    result: tuple[str, ...] | None = None

    if isinstance(source, str) and None not in seps and "" not in seps:
        segments: list[str | None] = []
        for sep in seps:
            pos: int = source.find(sep)
            if pos < 0:
                segments.append(None)
            else:
                segments.append(source[:pos])
                source = source[pos+len(sep):]
                if not source:
                    break

        segments.append(source)
        segments.extend([None] * (len(seps) - len(segments)))
        result = tuple(segments)

    return result


def str_to_lower(source: str) -> str:
    """
    Safely convert *source* to lower-case.

    If *source* is not a *str*, then it is itself returned.

    :param source: the string to convert to lower-case
    :return: *source* in lower-case, or *source* itself, if is not a string
    """
    return source.lower() if isinstance(source, str) else source


def str_to_upper(source: str) -> str:
    """
    Safely convert *source* to upper-case.

    If *source* is not a *str*, then it is itself returned.

    :param source: the string to convert to upper-case
    :return: *source* in upper-case, or *source* itself, if it is not a string
    """
    return source.upper() if isinstance(source, str) else source


def str_from_any(source: Any) -> str:
    """
    Convert *source* to its string representation.

    These are the string representations returned:
        - *None*: the string 'None'
        - *bool*: the string 'True' of 'Talse'
        - *str* : the source string itself
        - *bytes*: its hex representation
        - *date*: the date in ISO format (*datetime* is a *date* subtype)
        - *Path*: its POSIX form
        - all other types: their *str()* representation

    :param source: the data to be converted to string.
    :return: the string representation of the source data
    """
    # declare the return variable
    result: str

    # obtain the string representation
    if isinstance(source, bytes):
        result = source.hex()
    elif isinstance(source, date):
        result = source.isoformat()
    elif isinstance(source, Path):
        result = source.as_posix()
    else:
        result = str(source)

    return result


def str_to_bool(source: str) -> bool | None:
    """
    Obtain and return the *bool* value encoded in *source*.

    These are the criteria:
        - case is disregarded
        - the string values accepted to stand for *True* are *1*, *t*, or *true*
        - the string values accepted to stand for *False* are *0*, *f*, or *false*
        - all other values causes *None* to be returned

    :param source: the encoded bool value
    :return: the decoded bool value, or *None* if *source* fails the encoding criteria
    """
    # initialize the return variable
    result: bool | None = None

    if source in ["1", "t", "true"]:
        result = True
    elif source in ["0", "f", "false"]:
        result = False

    return result


def str_to_int(source: str) -> int | None:
    """
    Silently obtain and return the *int* value encoded in *source*.

    :param source: the encoded int value
    :return: the decoded *int* value, or *None* on error
    """
    # noinspection PyUnusedLocal
    result: int | None = None
    with suppress(Exception):
        result = int(source)

    return result


def str_to_float(source: str) -> float | None:
    """
    Silently obtain and return the *float* value encoded in *source*.

    :param source: the encoded float value
    :return: the decoded *float* value, or *None* on error
    """
    # noinspection PyUnusedLocal
    result: float | None = None
    with suppress(Exception):
        result = float(source)

    return result


def str_is_int(source: str) -> bool:
    """
    Determine whether *source* encodes a valid positive or negative integer.

    :param source: the encoded value
    :return: *True* if *source* encodes a valid integer, *False* otherwise
    """
    # declare the return variable
    result: bool

    try:
        int(source)
        result = True
    except ValueError:
        result = False

    return result


def str_is_float(source: str) -> bool:
    """
    Determine whether *source* encodes a valid positive or negative floating-point number.

    :param source: the encoded value
    :return: *True* if *source* encodes a valid floating-point number, *False* otherwise
    """
    # declare the return variable
    result: bool

    try:
        float(source)
        result = True
    except ValueError:
        result = False

    return result



import filetype
import puremagic
import mimetypes
from contextlib import suppress
from enum import StrEnum
from pathlib import Path
from tempfile import gettempdir
from typing import Final

from .env_pomes import APP_PREFIX, env_get_path

TEMP_FOLDER: Final[Path] = env_get_path(key=f"{APP_PREFIX}_TEMP_FOLDER",
                                        def_value=Path(gettempdir()))


# see https://mimetype.io/all-types
class Mimetype(StrEnum):
    """
    Commonly used mimetypes.
    """
    AAC = "audio/aac"
    AVI = " video/x-msvideo"
    BINARY = "application/octet-stream"
    BMP = "image/bmp"
    BZIP = "application/x-bzip"
    CSS = "text/css"
    CSV = "text/csv"
    DER = "application/x-x509-ca-cert"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    DWG = "image/vnd.dwg"
    EXE = "application/x-msdownload"
    FLAC = "audio/x-flac"
    FLV = "video/x-flv"
    GIF = "image/gif"
    GZIP = "application/gzip"
    HTML = "text/html"
    ICO = "image/x-icon"
    JAR = "application/java-archive"
    JAVASCRIPT = "text/javascript"
    JPEG = "image/jpeg"
    JPX = "image/jpx"
    JS = "text/javascript"
    JSON = "application/json"
    MKV = "video/x-matroska"
    MSWORD = "application/msword"
    MP3 = "audio/mpeg"
    MP4 = "video/mp4"
    MPEG = "video/mpeg"
    MIDI = "audio/midi"
    MULTIPART = "multipart/form-data"
    ODP = "application/vnd.oasis.opendocument.presentation"
    ODS = "application/vnd.oasis.opendocument.spreadsheet"
    OGG = "audio/ogg"
    P7B = "application/x-pkcs7-certificates"
    P7S = "application/pkcs7-signature"
    PDF = "application/pdf"
    PNG = "image/png"
    PPT = "application/vnd.ms-powerpoint"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    PSD = "image/vnd.adobe.photoshop"
    RAR = "application/x-rar-compressed"
    RTF = "application/rtf"
    SOAP = "application/soap+xml"
    SWF = "application/x-shockwave-flash"
    TEXT = "text/plain"
    TIFF = "image/tiff"
    URLENCODED = "application/x-www-form-urlencoded"
    WASM = "application/wasm"
    WAV = "audio/x-wav"
    WEBM = "audio/webm"
    WEBP = "image/webp"
    X7Z = "application/x-7z-compressed"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    XML = "application/xml"
    ZIP = "application/zip"


def file_get_data(file_data: Path | str | bytes,
                  max_len: int = None,
                  chunk_size: int = None) -> bytes | None:
    """
    Retrieve the data in *file_data*, or in a file in path *file_data*.

    The distinction is made with the parameter's type:
        - type *bytes*: *file_data* holds the data (returned as is)
        - type *str*: *file_data* holds the data (returned as utf8-encoded)
        - type *Path*: *file_data* is a path to a file holding the data

    :param file_data: the data as *bytes* or *str*, or the path to locate the file containing the data
    :param max_len: optional maximum length of the data to return, defaults to all data
    :param chunk_size: optional chunk size to use in reading the data, defaults to 128 KB
    :return: the data, or *None* if the file data could not be obtained
    """
    # initialize the return variable
    result: bytes | None = None

    # normalize the maximum length parameter
    if isinstance(max_len, bool) or \
       not isinstance(max_len, int) or max_len < 0:
        max_len = 0

    # normalize the chunk size
    if isinstance(chunk_size, bool) or \
       not isinstance(chunk_size, int) or chunk_size <= 0:
        chunk_size = 128 * 1024

    # what is the argument type ?
    if isinstance(file_data, bytes):
        # argument is type 'bytes'
        result = file_data

    elif isinstance(file_data, str):
        # argument is type 'str'
        result = file_data.encode()

    elif isinstance(file_data, Path):
        # argument is a file path
        file_bytes: bytearray = bytearray()
        file_path: Path = Path(file_data)
        # get the data
        with file_path.open(mode="rb") as f:
            buf_size: int = min(max_len, chunk_size) if max_len else chunk_size
            in_bytes: bytes = f.read(buf_size)
            while in_bytes:
                file_bytes += in_bytes
                if max_len:
                    if max_len <= len(file_bytes):
                        break
                    buf_size = min(max_len - len(file_bytes), chunk_size)
                else:
                    buf_size = chunk_size
                in_bytes = f.read(buf_size)
        result = bytes(file_bytes)

    if result and max_len and len(result) > max_len:
        result = result[:max_len]

    return result


def file_get_extension(mimetype: Mimetype | str) -> str | None:
    """
    Obtain and return the file extension best associated with mime type *mimetype*.

    This complements Python's *mimetypes.guess_extension()*, which fails for a number of
    frequently used mime types. In those cases, if *mimetype* is an instance of *Mimetype*,
    its *enum* name is used as extension. For consistency, the extension is returned as a
    lowercase string, with the leading dot ('.').

    :param mimetype: the reference mime type
    :return: the extension, with the leading dot ('.'), best associated with *mimetype*, or *None* on fail.
    """
    # initialize the return variable
    result: str = mimetypes.guess_extension(type=mimetype)

    if not result and mimetype in Mimetype:
        mimetype = Mimetype(mimetype)
        result = f".{mimetype._name_.lower()}"

    return result


def file_get_mimetype(file_data: Path | str | bytes) -> Mimetype | str:
    """
    Heuristics to determine the mimetype for *file_data*.

    The content is retrieved for analysis according to *file_data*'s type:
        - type *bytes*: *file_data* holds the data
        - type *str*: *file_data* holds the data as utf8-encoded
        - type *Path*: *file_data* is a path to a file holding the data

    The heuristics used, as heuristics go, provides an educated guess, not an accurate result.
    If a mimetype is found, and it is not in *Mimetype* (which is a small subset of known mimetypes),
    then its identifying string is returned.

    :param file_data: file data, or the path to locate the file
    :return: the probable mimetype, as a *Mimetype* object or as a string
    """
    mimetype: str | None = None
    if isinstance(file_data, str):
        file_data = file_data.encode()
    with suppress(TypeError):
        kind: filetype.Type = filetype.guess(obj=file_data)
        if kind:
            mimetype = kind.mime

    if not mimetype:
        if isinstance(file_data, Path):
            mimetype = puremagic.from_file(filename=file_data,
                                           mime=True)
        else:
            mimetype = puremagic.from_string(string=file_data,
                                             mime=True)
    result: Mimetype | str
    if mimetype:
        # for unknown mimetypes, return its identifying string
        result = Mimetype(mimetype) if mimetype in Mimetype else mimetype
    elif file_is_binary(file_data=file_data):
        result = Mimetype.BINARY
    else:
        result = Mimetype.TEXT

    return result


def file_is_binary(file_data: Path | str | bytes) -> bool:
    """
    Heuristics to determine whether the content of *file_data* is binary.

    The content is retrieved for analysis according to *file_data*'s type:
        - type *bytes*: *file_data* holds the data
        - type *str*: *file_data* holds the data as utf8-encoded
        - type *Path*: *file_data* is a path to a file holding the data

    The heuristics used, as heuristics go, provide an educated guess, not an accurate result.
    In the present case, the first 4 KBytes of the file data is inspected for the occurrence
    of characters normally absent in text files. Thus the presence of characters other than
    *bell*, *backspace*, *horizontal tab*, *newline*, *form feed*, *carriage return*, *escape*,
    and those in ASCII range [32 - 255] (except 127), would flag the file as binary.
    Empty or null content is considered to be non-binary.

    :param file_data: file data, or the path to locate the file
    :return: *True* if the determination resulted positive, *False* otherwise
    """
    # obtain up to 1024 bytes of content for analysis
    chunk: bytes = file_get_data(file_data=file_data,
                                 max_len=4096) or b""
    # check for null byte
    result: bool = b"\0" in chunk

    # check for non-printable characters
    if not result:
        # remove the chars listed below - chars remaining indicates content is binary
        #    7: \a (bell)
        #    8: \b (backspace)
        #    9: \t (horizontal tab)
        #   10: \n (newline)
        #   12: \f (form feed)
        #   13: \r (carriage return)
        #   27: \x1b (escape)
        #   0x20 - 0x100, less 0x7f: 32-255 char range, less 127 (the DEL control char)
        text_characters = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
        translation: bytes = chunk.translate(None,
                                             delete=text_characters)
        result = bool(translation)

    return result
